package c.l.blockchain.directory;

import leigh.mecha.log.MechaLogger;
import leigh.mecha.log.MechaLoggerFactory;

import java.time.Instant;
import java.time.temporal.ChronoUnit;

/**
 * Processes blockchain messages to build the state of the LeighChain directory.
 *
 * @author C. Alexander Leigh
 */
public class DirectoryStateMachine {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(DirectoryStateMachine.class);
    private final DirectoryStore store;

    public DirectoryStateMachine(DirectoryStore store) {
        this.store = store;
    }

    /**
     * Process a name claim on the behalf of an address.
     *
     * validityDays must be >=1 and <= 6*31. when is a trusted value.
     */
    public boolean processNameClaim(String address, String name,
                            int validityDays, String serviceUrl, Instant when) {

        if(validityDays > 6*31 || validityDays < 1) return false;

        Identity existing = store.getIdentityByName(name, when);

        if(existing!=null) {
            String existingAddress = existing.getAddress();

            if(!existingAddress.equals(address)) {
                logger.info("Name already claimed. [claimAddr: {}] [existingAddr: {}] [name: {}]",
                        address,existingAddress, name);
                return false;
            } else {
                // This claimant already has an existing valid claim. Just update the expiry time and service URL.
                // TODO: Set the serviceUrl here
                Instant newExpiry = when.plus(validityDays, ChronoUnit.DAYS);
                existing.setNameClaimExpire(newExpiry);
                existing.setServiceUrl(serviceUrl);
                return true;
            }
        }

        // No existing claim, so we straight up get to create one.
        Identity id = store.getIdentityByAddr(address);
        if(id==null) {
            id = new Identity();
            id.setAddress(address);
            id.setName(name);
            id.setServiceUrl(serviceUrl);
            store.putIdentity(id);
        }

        // TODO Set serviceUrl
        id.setNameClaimExpire(when.plus(validityDays, ChronoUnit.DAYS));

        return true;
    }

    @Override
    public String toString() {
        return "DirectoryStateMachine{" +
                "store=" + store +
                '}';
    }
}
